/*
  tidyup-mail - removes unsafe characters from email & makes xterm happy.
  Copyright (c) 2006 Alexander Gromnizki <iwerdon@gmail.com>.
  All rights reserved.

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is fur-
  nished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FIT-
  NESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

  $Id: main.c,v 1.10 2006/06/28 18:18:37 alex Exp $
*/

#include <stdio.h>
#include <stdbool.h>
#include <locale.h>
#include <unistd.h>
#include <assert.h>
#include <regex.h>

static char *id = "$Id: main.c,v 1.10 2006/06/28 18:18:37 alex Exp $";
static char *prog_ver = "0.12";

#define FILTER 1
#define MBOX 2
#define STATISTIC 4

struct chtab {
	int c;
	long count;
};

static size_t getline(FILE *, char **);
static void usage(const char *);
static void reverse(char *s);
static void itoa(const long num, char s[]);

/* works only for koi8-ru (russian, ukrainian, belarusian) locale */
static bool islegalchar(const int c)
{
	if (
		/* ASCII nonsense */
		(c >= 0 && c <= 8) ||
		(c >= 14 && c <= 31 ) ||
		c == 127 ||
		/* koi8-ru nonsense */
		(c >= 128 && c <= 162) ||
		c == 165 ||
		(c >= 168 && c <= 172) ||
		(c >= 175 && c <= 178) ||
		c == 181 ||
		(c >= 184 && c <= 188) ||
		c == 191
		) return false;
	else return true;
}

static int ch_exists(const struct chtab bullshit[], const int n, const int c)
{
	int pos=n;
	while (pos-- > 0) if ((bullshit++)->c == c) return pos;
	return -1;
}

static bool chars_info(const int c, long *ok,
					   struct chtab bullshit[], int *bn, long *bstotal)
{
	int pos;

	if (islegalchar(c)) {
		++*ok;
		return true;
	}

    /* this can't happens, but just in case, you know... */
	if (*bn > 256) return false;

	if ((pos = ch_exists(bullshit, *bn, c)) == -1) {
		bullshit[*bn].c = c;
		bullshit[*bn].count=1;
		++*bn;
	} else bullshit[pos].count++;

	++*bstotal;
	return false;
}

static void line_info(const unsigned char *line, char **trline,
					  long *ok, struct chtab bullshit[], int *bn,
					  long *bstotal, const int flags)
{
	bool r;
	char *p;

	*trline = (char*)realloc(*trline, strlen(line)+1);
	assert(*trline);
	p = *trline;

	while (*line) {
		if (flags & STATISTIC) r = chars_info(*line, ok,
											  bullshit, bn, bstotal);
		else r = islegalchar(*line);
		if (r) *p++ = *line;
		line++;
	}
	*p = '\0';
}

static void print_stat(const long ok,
					   const struct chtab bullshit[], const int bn,
					   const long bstotal)
{
	int len = bn;
	int col=0;
	char ts[32], s[32];
		
	fprintf(stderr, "Total legal characters: %d\n"
			"Total bullshit: %d\n", ok, bstotal);

	if (len) {
		fprintf(stderr, "Which consists of (character/count):\n");
		while (len-- > 0) {
			if (col == 5) {
				fprintf(stderr, "\n");
				col=0;
			}
			
			itoa(bullshit->c, ts);
			strcpy(s, ts);
			itoa(bullshit->count, ts);
			strcat(s, "/");
			strcat(s, ts);
			
			fprintf(stderr, (len > 0) ? "%10s " : "%10s\n", s);
			++bullshit;
			++col;
		}
	} else fprintf(stderr, "You got lucky.\n");
}

int main(int argc, char *argv[])
{
	int c;
	long ok=0;					/* total number of legal chars*/
	char *line = NULL,			/* current line from the stream */
		*trline = NULL;			/* current line from the stream without
								   bad chars */
	size_t len;					/* current line length */
	FILE *fp;
	bool reading_fstdin=false;
	regex_t mbox_postmark;
	/* pointer to the array of bad chars structure */
	struct chtab bullshit[256];
	/* number of unique bad chars (and array index too) */
	int bn=0;
	long bstotal=0;				/* total number of bad chars */

	/* options */
	int flags = 0;
	bool only_stat = false;

	setlocale(LC_ALL, "");

	flags |= FILTER;
	while ((c = getopt(argc, argv, "hsSm")) != -1)
		switch (c) {
	     case 'S':
			 only_stat = true;
	     case 's':
			 flags |= STATISTIC;
		     break;
	     case 'm':
			 flags |= MBOX;		/* stream is mbox message format */
		     break;
	     case '?':
	     case 'h':
	     default:
		     usage(argv[0]);
			 exit(1);
			 break;
		}
	optind--;
	argc -= optind;
	argv += optind;

	if (argc <= 1) {
		fp = stdin;
		reading_fstdin = true;
	}

	if (only_stat) {			/* remove any other options */
		flags &= ~FILTER;
		flags &= ~MBOX;
	}

#ifdef DEBUG
	if (flags & FILTER) fprintf(stderr, "FILTER\n");
	if (flags & STATISTIC) fprintf(stderr, "STATISTIC\n");
	if (flags & MBOX) fprintf(stderr, "MBOX\n");
/*	exit(0);*/
#endif

	/* RE to match mbox postmark line */
	if (0 != regcomp(&mbox_postmark, "^From [^[:space:]]*@[^[:space:]]*  ?"
					 "(Mon|Tue|Wed|Thu|Fri|Sat|Sun) "
					 "(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)  ?"
					 "[0-9]+ [0-9]{2}:[0-9]{2}:[0-9]{2} "
					 "([-+][0-9]{4})? ?[0-9]{4}[[:space:]]*$",
					 REG_EXTENDED|REG_NOSUB))
		err(2, "couldn't compile RE");

	/*
	  Examing files or stdin for input
	*/
	while (true) {
		bool mbox_message=false; /* is now a lines of mail envelope */

		if (!reading_fstdin) {
			if (--argc == 0) break;	/* no more files */
			if ((fp = fopen(*++argv, "r")) == NULL)
				err(1, "%s", *argv);
		}

		/* reading stream */
		while ((len = getline(fp, &line)) > 0) {
			line_info(line, &trline, &ok,
					  bullshit, &bn, &bstotal, flags);

			if (flags & MBOX) {
				/* insert header at the end of the envelope */
				if (!mbox_message &&
					0 == regexec(&mbox_postmark,
								 trline, (size_t)0, NULL, 0))
					mbox_message = true;
				else if (mbox_message && strcmp(trline, "\n") == 0) {
					printf("X-TidyUp-Mail: yes\n");
					mbox_message = false;
				}

				printf("%s", trline);
			} else if (flags & FILTER) printf("%s", trline);
		}

		if (reading_fstdin) break;
		else fclose(fp);
	}

	free(line);
	free(trline);
	regfree(&mbox_postmark);

	/*
	  Print statistics
	*/
	if (flags & STATISTIC) print_stat(ok, bullshit, bn, bstotal);

	return 0;
}

static void usage(const char *name)
{
	printf("tidyup-mail %s - removes unsafe characters from email & makes xterm happy.\n"
		   "Copyright (c) 2006 Alexander Gromnizki <iwerdon@gmail.com>. "
	  	   "All rights reserved.\n\n"
		   "Usage: %s [-msSh] [file(s) ...]\n"
		   "        -s      show statistics on stderr after filtering.\n"
		   "        -m      suppose that file or stdin is mbox mail.\n"
		   "                storage, so add at the end of the envelope\n"
		   "                a mark header.\n"
		   "        -S      show ONLY statistics (on stderr).\n"
		   "        -h      this help.\n"
		   "\n"
		   "Examples:\n"
		   "%% tidyup-mail one.txt two.txt > three.txt\n"
		   "%% cat archive.mbox | tidyup-mail -S\n",
		   prog_ver, name);
}

static size_t getline(FILE *stream, char **line)
{
	size_t c, len=0, size=BUFSIZ;
	char *p;

	*line = (char*)realloc(*line, size);
	assert(*line);
	p = *line;
	while ((c = getc(stream)) != EOF && c != '\n') {
		if (len == size) {
			size *= 2;
			*line = (char*)realloc(*line, size);
			assert(*line);
			p = *line+len;
		}
		*p++ = c;
		++len;
	}

	if (c == '\n') {
		*p++ = c;
		++len;
	}
	*p = '\0';
	return (size_t)(p-*line);
}

static void reverse(char *s)
{
	char t;
	char *p = s+strlen(s)-1;	/* points to the last char */

	while (s < p) {
		t = *s;
		*s++ = *p;
		*p-- = t;
	}
}

static void itoa(const long num, char s[])
{
	long n=num, i = 0, ngtv = 0;
	char *ps = s;
	
	if (n < 0) {
		ngtv=1;
		n = -n;
	}

	while (n) {
		*ps++ = n%10 + '0';
		n /= 10;
	}
	if (ngtv) *ps++ = '-';
	*ps = '\0';
	
	reverse(s);
}
